#include <Redis_Receiver.h>

#include <sys/socket.h>
#include <arpa/inet.h>
#include <fstream>
#include <map>
#include <string.h>
#include <sstream>
#include <regex>

#include <iostream>
#include <thread>
#include <chrono>
#include <ctime>
#include <iomanip>

using namespace inaf::oasbo::Receivers;

RedisReceiver::RedisReceiver() :
		RedisReceiver("127.0.0.1", 6379, "Astri_ADAS_key") {
}

RedisReceiver::RedisReceiver(std::string ip, int prt, std::string key) :
		ip(ip), port(prt), key(key) {
	host = ip + ":" + std::to_string(port) + "/" + key;
	this->check_interval = 1;
}

RedisReceiver::~RedisReceiver() {
	closeConnectionToClient();
}

int RedisReceiver::connectToClient() {
	if (isConnectedToClient())
		return 1;
	context = redisConnect(ip.c_str(), port);
	if (context == nullptr || context->err) {
		if (context) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Receiver]\t" << context->errstr << std::endl;
		} else {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Receiver]\t"
					<< " Can't allocate Redis context" << std::endl;
		}
		return -1;
	}
	time_t now = time(nullptr);
	std::cout << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
			<< "]\t[Redis Receiver]\t" << "Connected" << std::endl;
	return 1;
}

int RedisReceiver::closeConnectionToClient() {
	if (isConnectedToClient()) {
		redisFree(this->context);
		context = nullptr;
	}
	return 1;
}

bool RedisReceiver::isConnectedToClient() const {
	return !(context == nullptr || context->err);
}

int RedisReceiver::receiveFromClient(Packets::BasePacket &pack) {
	if (!isConnectedToClient()) {
		if (context) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Receiver]\t" << context->errstr << std::endl;
		} else {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Receiver]\t" << "Can't allocate Redis context"
					<< std::endl;
		}
		return -1;
	}

	std::string cmd("RPOP ");
	cmd.append(key);
	redisReply *r = (redisReply*) redisCommand(context, cmd.c_str());
	if (r == nullptr) {
		std::this_thread::sleep_for(std::chrono::seconds(this->check_interval));
		return -1;
	}
	if (r->type == REDIS_REPLY_NIL) {
		freeReplyObject(r);
		std::this_thread::sleep_for(std::chrono::seconds(this->check_interval));
		return -1;
	}
	if (r->type == REDIS_REPLY_ERROR) {
		freeReplyObject(r);
		std::this_thread::sleep_for(std::chrono::seconds(this->check_interval));
		return -1;
	}
	if (r->type == REDIS_REPLY_STRING && r->type != REDIS_REPLY_NIL) { // Received
		int size = r->len;
		ssize_t header_size = pack.getHeaderSize();
		if (size < header_size) {
			freeReplyObject(r);
			return -1;
		}
		pack.copyToMemory(reinterpret_cast<uint8_t*>(r->str), header_size);
		ssize_t tot_packet_size = pack.getHeaderSize() + pack.getPayloadSize()
				+ pack.getTailSize();

		if (!pack.hasRecognizedHeader() || size < (tot_packet_size)) {
			time_t now = time(nullptr);
			std::cerr << "["
					<< std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
					<< "]\t[Redis Receiver]\t" << "packet not recognized: "
					<< pack["type"].value() << " " << pack["subtype"].value()
					<< " " << pack.getPayloadSize() << std::endl;
			freeReplyObject(r);
			resetPacket(pack, header_size);
			return -1;
		}

		pack.copyToMemory(reinterpret_cast<uint8_t*>(r->str), size);
		freeReplyObject(r);
		return size;
	}
	// other errors
	freeReplyObject(r);
	return -1;
}

std::string RedisReceiver::getHost() {
	return ip + ":" + std::to_string(port) + "/" + key;
}

void RedisReceiver::setHost(std::string host) {
	std::string ip = { };
	int port = 0;
	std::string key = { };
	if (split_ip_port_key(host, ip, port, key)) {
		setIp(ip);
		setPort(port);
		setKey(key);
	} else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Redis Receiver]\t" << host
				<< "\" does not match the format: "
				<< "\"xxx.xxx.xxx.xxx:xxxx/xxxx\"" << std::endl;
	}
}

void RedisReceiver::setIp(std::string ip) {
	std::regex pattern("^(\\d{1,3}\\.){3}\\d{1,3}$");
	// Check if the input string matches the expected format
	if (std::regex_match(ip, pattern)) {
		this->ip = ip;
		this->host = ip + ":" + std::to_string(port) + "/" + key;
	} else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Redis Receiver]\t" << "invalid IP address: " << ip
				<< std::endl;
	}
}

void RedisReceiver::setPort(int port) {
	if (port > 1023 && port < 65535) {
		this->port = port;
		this->host = ip + ":" + std::to_string(port) + "/" + key;
	} else {
		time_t now = time(nullptr);
		std::cerr << "[" << std::put_time(localtime(&now), "%Y-%m-%d %H:%M:%S")
				<< "]\t[Redis Receiver]\t" << "Invalid port : " << port
				<< std::endl;
	}
}

void RedisReceiver::setKey(std::string key) {
	this->key = key;
	this->host = ip + ":" + std::to_string(port) + "/" + key;
}

bool RedisReceiver::split_ip_port_key(const std::string &ip_port_key,
		std::string &ip_address, int &port, std::string &key) {
	// Regex pattern to match IP address and port in the format "xxx.xxx.xxx.xxx:xxxx/xxxx"
	std::regex pattern("^(\\d{1,3}\\.){3}\\d{1,3}:(\\d{1,5})/(\\w+)$");
	std::smatch matches;

	// Check if the input string matches the expected format
	if (!std::regex_match(ip_port_key, matches, pattern)) {
		return false;
	}
	ip_address = matches.str(1);
	port = std::stoi(matches.str(2));
	key = matches.str(3);
	return true;
}

void RedisReceiver::resetPacket(Packets::BasePacket &pack, int bytes) {
	uint8_t *buff = new uint8_t[bytes];
	std::memset(buff, 0, bytes);
	int toBeReset = std::min(
			static_cast<int>(pack.getPacketStructureByteSize()), bytes);
	pack.copyToMemory(buff, toBeReset);
	delete[] buff;
}
